#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#include <lauxlib.h>
#include "lb-messages.h"
#include "lblua-utils.h"

void
lblua_checkargcount(lua_State* ls, int count, const char* function)
{
	lb_assert(ls != NULL);
	lb_assert(function != NULL);

	if (lua_gettop(ls) != count)
		luaL_error(ls, "wrong number of arguments to `%s'", function);
}

int
lblua_range_error(lua_State* ls, const char* function)
{
	lb_assert(ls != NULL);
	lb_assert(function != NULL);

	return luaL_error(ls, "range error in `%s'", function);
}

void
lblua_push_array_int(lua_State* ls, int count, const int* data)
{
	int i;

	lb_assert(ls != NULL);

	lua_newtable(ls);
	luaL_setn(ls, -1, count);

	for (i = 1; count > 0; count--) {
		lua_pushnumber(ls, *data++);
		lua_rawseti(ls, -2, i++);
	}
}

int
lblua_check_array_int(lua_State* ls, int idx, int bsize, int* buf)
{
	lb_assert(ls != NULL);
	lb_assert(bsize > 0 && buf);

	if (!lua_istable(ls, idx))
		luaL_error(ls, "`table' expected, got `%s'",
			       lua_typename(ls, idx));
	else {
		int i, n = luaL_getn(ls, idx);

		n = LB_MIN(n, bsize);

		for (i = 0; i < n; ++i) {
			lua_rawgeti(ls, idx, i + 1);

			buf[i] = (int) lua_tonumber(ls, -1);
			if (buf[i] == 0 && !lua_isnumber(ls, -1))
				luaL_error(ls, "array of numbers expected");

			lua_pop(ls, 1);
		}

		return n;
	}

	return 0;
}

void
lblua_push_array_double(lua_State* ls, int count, const double* data)
{
	int i;

	lb_assert(ls != NULL);

	lua_newtable(ls);
	luaL_setn(ls, -1, count);

	for (i = 1; count > 0; count--) {
		lua_pushnumber(ls, *data++);
		lua_rawseti(ls, -2, i++);
	}
}

int
lblua_check_array_double(lua_State* ls, int idx, int bsize, double* buf)
{
	lb_assert(ls != NULL);
	lb_assert(bsize > 0 && buf);

	if (!lua_istable(ls, idx))
		luaL_error(ls, "`table' expected, got `%s'",
			       lua_typename(ls, idx));
	else {
		int i, n = luaL_getn(ls, idx);

		n = LB_MIN(n, bsize);

		for (i = 0; i < n; ++i) {
			lua_rawgeti(ls, idx, i + 1);

			buf[i] = (double) lua_tonumber(ls, -1);
			if (buf[i] == 0 && !lua_isnumber(ls, -1))
				luaL_error(ls, "array of numbers expected");

			lua_pop(ls, 1);
		}

		return n;
	}

	return 0;
}

void
lblua_push_simple_table(lua_State* ls, int count,
			  const char* keys[], const double vals[])
{
	lb_assert(ls != NULL);
	lb_assert(count == 0 || (keys && vals));

	lua_newtable(ls);
	for (; count > 0; --count) {
		lua_pushstring(ls, *keys++);
		lua_pushnumber(ls, *vals++);
		lua_rawset(ls, -3);
	}
}

void
lblua_check_simple_table(lua_State* ls, int idx, int count,
			   const char* keys[], double vals[])
{
	lb_assert(ls != NULL);
	lb_assert(count == 0 || (keys && vals));

	if (!lua_istable(ls, idx))
		luaL_error(ls, "`table' expected, got `%s'",
			       lua_typename(ls, idx));
	else {
		int i;

		for (i = 0; i < count; ++i) {
			lua_pushstring(ls, keys[i]);
			lua_gettable(ls, idx);

			vals[i] = lua_tonumber(ls, -1);
			if (vals[i] == 0 && !lua_isnumber(ls, -1))
				luaL_error(ls, "table element '%s' should"
					       " be of type 'number' instead"
					       " of '%s'", keys[i],
					       lua_typename(ls, -1));

			lua_pop(ls, 1);
		}
	}
}
