--
--  parameters
--

NX = 128
NY = 64

PARAMETERS = {
	T = 0.56,
	a = 9/49,
	b = 2/21,
	K = 0.01,
	tau = 0.8
}

PY = false -- has walls in Y directions

VT = 0.05 -- velocity of the top wall
VB = -0.05 -- velocity of the bottom wall

START = 0 -- start time
END = 2048 -- end time
FREQ = 64 -- how frequently density snapshots should be rendered

--
--  initializer
--

initializer = function(x, y)
	return 2.5 + math.random(), 0, 0
end

--
--  density to color mapping
--

cmap = lb.colormap()
cmap:append_color(1, 0, 0)
cmap:append_color(1, 1, 1)
cmap:append_color(0, 0, 1)

density2rgb = function(rho, ux, uy)
	return cmap:map_value((rho - 2.7)/3)
end

--
--  initializes averages
--

initialize = function(simulation, initializer)
	local pinfo = simulation:partition_info()
	local nx, ny = pinfo.size[1], pinfo.size[2]
	local x0, y0 = pinfo.global_origin[1], pinfo.global_origin[2]

	for x = 0, nx - 1 do
		for y = 0, ny - 1 do
			simulation:set_averages(x, y,
						initializer(x + x0, y + y0))
		end
	end

	simulation:set_equilibrium()
end

--
--  each process creates its own RGB buffer; master glues them together
--

render1 = function(callback, simulation, filename)
	local pinfo = simulation:partition_info()
	local nx, ny = pinfo.size[1], pinfo.size[2]
	local rgb = lb.rgb(nx, ny)

	for x = 0, nx - 1 do
		for y = 0, ny - 1 do
			local rho, ux, uy = simulation:get_averages(x, y)
			local R, G, B = callback(rho, ux, uy)
			rgb:set_pixel(x, ny - y - 1, R, G, B)
		end
	end

	rgb:save(filename)
	collectgarbage() -- this should on C side
end

filename_IJ = function(base, i, j)
	return string.format("%s-%d-%d.png", base, i, j)
end

renderN = function(callback, simulation, filename)
	local pinfo = simulation:partition_info()

	render1(callback, simulation,
		filename_IJ(filename, pinfo.processor_coords[1],
		    pinfo.processors_size[2] - pinfo.processor_coords[2] - 1))
	lb.mpi_barrier()

	if pinfo.processor_rank == 0 then
		local cmd, filez
		cmd = string.format("montage +frame +shadow +label" ..
				    " -tile %dx%d -geometry %dx%d+0+0",
				    pinfo.processors_size[1],
				    pinfo.processors_size[2],
				    pinfo.size[1], pinfo.size[2])

		filez = ""
		for j = 0, pinfo.processors_size[2] - 1 do
			for i = 0, pinfo.processors_size[1] - 1 do
				filez = filez .. " "
				    .. filename_IJ(filename, i, j)
			end
		end

		cmd = cmd .. " " .. filez .. " -depth 8 " .. filename
--		print(cmd)
		os.execute(cmd)

		cmd = "rm -f " .. filez
--		print(cmd)
		os.execute(cmd)
	end
end

make_filename = function(t)
	local res = ""
	local digits = 1

	for n = 1, 8 do
		pp = __pow(10, n)
		if math.floor(t / pp) == 0 then
			break
		else
			digits = digits + 1
		end
	end

	todo = 7 - digits
	for n = 0, todo do
		res = res .. "0"
	end

	return res .. t .. ".png"
end

render = function(callback, simulation, t)
	local filename = make_filename(t)

	if lb.is_parallel() then
		renderN(callback, simulation, filename)
	else
		render1(callback, simulation, filename)
	end
end

--
--  usefull function
--

report_progress = function(start_wtime, t0, t1, t)
	local secs = lb.wtime() - start_wtime
	local steps = t - t0
	local sps = secs/steps

	print()
	io.write(string.format("<>> %d steps done in %f seconds (%f sec/step)\n",
			       steps, secs, sps))
	io.write(string.format("<>> ETA: %s\n", lb.wtime_string((t1 - t)*sps)))
end

--
--  finally ... core functionality
--

simulation = lb.d2q9_VdW(NX, NY, PY)
pinfo = simulation:partition_info()
math.randomseed(pinfo.processor_rank + 1)

simulation:set_parameters(PARAMETERS)
if not PY then
	simulation:set_walls_speed(VT, VB)
end

initialize(simulation, initializer)

t0 = lb.wtime()
last_report = t0

for t = START, END do
	simulation:advance()

	if pinfo.processor_rank == 0 then
		local now = lb.wtime()
		if now - last_report > 7 then
			report_progress(t0, START, END, t)
			last_report = now
		end
	end

	if math.mod(t, FREQ) == 0 then
		render(density2rgb, simulation, t)
	end
end
